
// Billboard.java
// Andrew Davison, dandrew@fivedots.coe.psu.ac.th, November 2004

/* The Billboard is a square of length size, resting on the XZ plane,
   centered at (x,z). Wrapped over it is an image. Any alpha parts
   of the image will not be rendered.

   The billboard mesh is a triangle array defining the square.

   The billboard is aligned (made to face) the camera when its
   align() method is called. The alignment is along the billboard's
   z-axis.
*/

import javax.microedition.lcdui.*;
import javax.microedition.m3g.*;


public class Billboard 
{
  private Mesh bbMesh;
  private Group cameraGroup;   
     // the billboard aligns itself with this camera position


  public Billboard(Image2D bbImage, Group camGroup, float x, float z, float size) 
  { 
    cameraGroup = camGroup;

    // build bbMesh
    VertexBuffer bbVertBuf = makeGeometry();

    int[] indicies = {1,2,0,3};  // the billboard is one quad
    int[] stripLens = {4};
    IndexBuffer bbIdxBuf = new TriangleStripArray(indicies, stripLens);

    Appearance bbApp = makeAppearance(bbImage);

    bbMesh = new Mesh(bbVertBuf, bbIdxBuf, bbApp);

    float size2 = size * 0.5f;   
       /* The mesh is 2-by-2 in size, and so the extra 0.5 factor
          in the scaling reduces it to 1-by-1. */
    bbMesh.scale(size2, size2, size2);
    bbMesh.setTranslation(x, size2, z);

    bbMesh.setAlignment(cameraGroup, Node.Z_AXIS, null, Node.NONE);  
         /* The billboard alignment will be along its z-axis only,
            no y-axis alignment is employed. */
  }  // end of Billboard()


  private VertexBuffer makeGeometry()
  /*  The geometry defines a square resting on top of the XZ plane,
      centered at (0,0), with sides of length 2.
      There are no normals, but there are texture coords.
  */
  {
    /* Create vertices, starting at the bottom left and going 
       counter-clockwise. */
    short[] verts = {-1,-1,0,  1,-1,0,  1,1,0,  -1,1,0};
    VertexArray va = new VertexArray(verts.length/3, 3, 2);
    va.set(0, verts.length/3, verts);

    // create texture coordinates using the same order as the vertices
    short[] tcs = {0,1,  1,1,  1,0,  0,0};
    VertexArray texArray = new VertexArray(tcs.length/2, 2, 2);
    texArray.set(0, tcs.length/2, tcs);

    VertexBuffer vb = new VertexBuffer(); 
    vb.setPositions(va, 1.0f, null); // no size, bias
    vb.setTexCoords(0, texArray, 1.0f, null);

    return vb;
  }  // end of makeGeometry()


  private Appearance makeAppearance(Image2D bbImage)
  /* The image's alpha component will cause the square's
     surface to be invisible at those locations.
  */
  { Appearance app = new Appearance();

    // The alpha component of the texture will be used in the rendering
    CompositingMode compMode = new CompositingMode();
    compMode.setBlending(CompositingMode.ALPHA);
    app.setCompositingMode(compMode);

    if (bbImage != null) {
      Texture2D tex = new Texture2D(bbImage);
      tex.setFiltering(Texture2D.FILTER_NEAREST, Texture2D.FILTER_NEAREST);
      tex.setWrapping(Texture2D.WRAP_CLAMP, Texture2D.WRAP_CLAMP);
      tex.setBlending(Texture2D.FUNC_REPLACE); 
                    // the texture with replace the surface

      app.setTexture(0, tex);
    }
    return app;
  }  // end of makeAppearance()


  // -------------------- access method ----------------------

  public Mesh getBillboardMesh()
  {  return bbMesh;  }


  public void align()
  // align the board's z-axis with the current position of the cameraGroup node
  { bbMesh.align(cameraGroup);  }

} // end of Billboard Class
