
// SkinCanvas.java
// Andrew Davison, dandrew@fivedots.coe.psu.ac.th, November 2004

/* Create a 3D scene in retained mode (i.e. build a scenegraph).

   Scene elements: a bending block, grassy floor, trees, a
   moving sky background, a moveable camera.

   The scene is updated and repainted by a timer task in 
   SkinM3G which calls update() periodically.
*/

import javax.microedition.lcdui.*;
import javax.microedition.m3g.*;


public class SkinCanvas extends Canvas implements CommandListener 
{
  private Command exitCmd; 

  private Graphics3D iG3D;

  private SkinM3G top;
  private World scene;

  // scene elements
  private MobileBackground mobileBackGnd;
  private MobileCamera mobCam;
  private Group cameraGroup;
  private Billboard bb1, bb2, bb3;   // tree billboards

  // timing information
  private int appTime = 0;
  private int nextTimeToAnimate;


  public SkinCanvas(SkinM3G top) 
  { 
    this.top = top;

    exitCmd = new Command("Exit", Command.EXIT, 0);
    addCommand(exitCmd);
    setCommandListener(this);

    iG3D = Graphics3D.getInstance();
    scene = new World();   // create scene graph
    buildScene();

    // start the animation
    nextTimeToAnimate = scene.animate(appTime);
  }  // end of SkinCanvas()


  public void commandAction(Command command, Displayable disp)
  { 
    if (command == exitCmd)    // finish the morphing
      top.finishGame();
    else   // execution should not get here
      System.out.println("Command Action Error: " + command);
  }


  private void buildScene()
  /* Add a background, camera, trees, the bending model,
     and a floor to the scene. There are no lights needed,
     since nothing uses a Material component.
  */
  { /* The ordering of the first three methods is important.
       The Billboard trees need a reference to the camera, and
       the camera needs a reference to the background. */
    addBackground(); 
    addCamera();
    addTrees();

    SkinModel sm = new SkinModel();
    scene.addChild( sm.getModel() );  // add the model

    addFloor();
  } // end of buildScene()


  private void addBackground()
  // the background is a MobileBackground object
  {    
    Image2D backIm = loadImage("/clouds.gif");
    mobileBackGnd = new MobileBackground(backIm, getWidth(), getHeight());
    scene.setBackground( mobileBackGnd.getBackground() );
  }


  private void addCamera()
  {
    mobCam = new MobileCamera( getWidth(), getHeight(), mobileBackGnd);  
     /* The camera needs a reference to the MobileBackground
        object, so it can tell it to move. */

    cameraGroup = mobCam.getCameraGroup();
    scene.addChild( cameraGroup );
    scene.setActiveCamera( mobCam.getCamera() ); 
  }  // end of addCamera()


  private void addTrees()
  /* Create two kinds of trees, and add them to the scene:
         ts1, ts2: Sprite3D objects;
         bb1, bb2, bb3: Billboard objects
  */
  { 
    Sprite3D ts1 = makeSprite("/tree1.gif", 1.5f, 0, 1.0f);
    scene.addChild( ts1 );

    Sprite3D ts2 = makeSprite("/tree2.gif", -1.5f, 0, 1.5f);
    scene.addChild( ts2 );

    Image2D bbIm1 = loadImage("/tree2.gif");
    bb1 = new Billboard(bbIm1, cameraGroup, 2.0f, -1.0f, 2.0f);
                            // needs a ref. to cameraGroup
    scene.addChild( bb1.getBillboardMesh() );

    Image2D bbIm2 = loadImage("/tree3.gif");
    bb2 = new Billboard(bbIm2, cameraGroup, -2.0f, -2.0f, 1.5f);
                            // needs a ref. to cameraGroup
    scene.addChild( bb2.getBillboardMesh() );

    Image2D bbIm3 = loadImage("/tree4.gif");
    bb3 = new Billboard(bbIm3, cameraGroup, 0.5f, -3.2f, 1.5f);
                            // needs a ref. to cameraGroup
    scene.addChild( bb3.getBillboardMesh() );
  } // end of addTrees()


  private Sprite3D makeSprite(String treeFnm, float x, float z, float size) 
  /* Create a sprite resting on the floor, centered at (x,z), with
     sides of size length. Alpha parts of the texture are not displayed.
  */
  { 
    Image2D image2D = loadImage(treeFnm);
    // System.out.println(treeFnm + " format: " + image2D.getFormat());

    // don't display the alpha parts of the image
    Appearance app = new Appearance();
    CompositingMode compMode = new CompositingMode();
    compMode.setBlending(CompositingMode.ALPHA);
    app.setCompositingMode(compMode);

    Sprite3D sprite = new Sprite3D(true, image2D, app);
       // create a unit-size sprite

    sprite.scale(size, size, size);  // make the sprite size in length
    sprite.setTranslation(x, (0.5f * size), z);  
             // move it up to rest on the floor

    return sprite;
  }  // end of makeSprite()


  private void addFloor()
  /* We have two floor-making choices. Either the Floor class, first
     used in AnimM3G, which creates one big square. Or TiledFloor which
     covers the same area with 1-by-1 tiles. Each tile is individually
     covered with the texture, which makes the overall pattern look better. 
     Try comparing the two classes using the texture in grass.gif
  */
  {
    // Image2D floorIm = loadImage("/grass.gif"); 
    // Floor f = new Floor( floorIm, 8);   
          // 8 by 8 size, made from one big square

    Image2D floorIm = loadImage("/grass.gif"); 
    TiledFloor f = new TiledFloor( floorIm, 8);   
      // 8 by 8 size, made up of 1 by 1 tiles

    scene.addChild( f.getFloorMesh() );  // add the floor
  }  // end of addFloor()


  private Image2D loadImage(String fn)
  // load the image stored in fn
  { Image2D im = null;
    try {
       im = (Image2D)Loader.load(fn)[0];
    }
    catch (Exception e)
    { System.out.println("Cannot make image from " + fn); }
    return im;
  }  // end of loadImage()



  // ------------ process key presses and releases ------------


  protected void keyPressed(int keyCode)
  { int gameAction = getGameAction(keyCode);
    mobCam.pressedKey(gameAction);
  }

  protected void keyReleased(int keyCode)
  { int gameAction = getGameAction(keyCode);
    mobCam.releasedKey(gameAction);
  }

  // ------------------ update and paint the canvas ------------------

  public void update()
  // called by the timer task in SkinM3G
  {
    mobCam.update();

    bb1.align();  // align the billboards with the camera position
    bb2.align();
    bb3.align();

    appTime++;
    if (appTime >= nextTimeToAnimate)
      nextTimeToAnimate = scene.animate(appTime) + appTime;
    repaint();
  }  // end of update()


  protected void paint(Graphics g)
  {
    iG3D.bindTarget(g);
    try {
      iG3D.render(scene);
    }
    catch(Exception e)
    { e.printStackTrace();  }
    finally {
      iG3D.releaseTarget();
    }

    // show the camera's details
    g.drawString( mobCam.getKeyMode(), 5,5, Graphics.TOP|Graphics.LEFT);
    g.drawString( mobCam.getPosition(), 5,18, Graphics.TOP|Graphics.LEFT);
    g.drawString( mobCam.getRotation(), 5,31, Graphics.TOP|Graphics.LEFT);
  } // end of paint()


}  // end of SkinCanvas class
