
// SkinModel.java
// Andrew Davison, dandrew@fivedots.coe.psu.ac.th, November 2004

/* The skinned mesh is a block with coloured vertices.

   It rests on the XZ plane, centered at (0,0). It contains three
   joints, at its base, middle and top. Each joint is linked to the
   four vertices at that level in the block.

   Animation is used to rotate the joints in the base and the middle
   around the z-axis.
*/


import javax.microedition.lcdui.*;
import javax.microedition.m3g.*;


public class SkinModel 
{
  private SkinnedMesh blockModel;
  private Group startJoint, midJoint, endJoint;
  private AnimationController animController;


  public SkinModel() 
  { makeBlock();
    setUpAnimation();
  }

  public SkinnedMesh getModel()
  {  return blockModel;  }


  // ------------------- model creation -----------------------------

  private void makeBlock()
  /* The block has coloured vertices, but no normals or texture coordinates. 
     It has 3 joints attached to it.
  */
  { // create vertices
    short[] verts = getVerts();
    VertexArray va = new VertexArray(verts.length/3, 3, 2);
    va.set(0, verts.length/3, verts);

	// create vertex colours
    byte[] cols = getColours();
	VertexArray colsArray = new VertexArray(cols.length/3, 3, 1);
	colsArray.set(0, cols.length/3, cols);

    VertexBuffer vb = new VertexBuffer(); 
    vb.setPositions(va, 1.0f, null); // no scale, bias
	// vb.setDefaultColor(0x00FF0000); // default to red vertices
	vb.setColors(colsArray);

    IndexBuffer ib = 
          new TriangleStripArray(getStripIndicies(), getStripLengths());

    Appearance app =  new Appearance();   // no material

    // create skinned mesh
    blockModel = new SkinnedMesh(vb, ib, app, makeSkeleton());
    blockModel.scale(0.1f, 0.1f, 0.1f);

    /* Link joints to the block's vertices. Each joint is linked to
       the four vertices at that level in the block. */
    blockModel.addTransform(startJoint, 1, 0, 4); 
    blockModel.addTransform(midJoint, 1, 4, 4);
    blockModel.addTransform(endJoint, 1, 8, 4);
  }  // end of makeBlock()


  private Group makeSkeleton()
  /* The skeleton is a chain of three joints (Group nodes).
     start --> mid --> end

     The start and mid joints will be rotated.

     The start joint will be linked to the bottom of the 
     block, the middle joint to the block's middle,
     and the top joint to the block's top. 
  */
  { Group skeleton = new Group();

    startJoint = new Group();
    skeleton.addChild(startJoint);

    midJoint = new Group();
    midJoint.setTranslation(0.0f, 4.0f, 0.0f);
    startJoint.addChild(midJoint);

    endJoint= new Group();
    endJoint.setTranslation(0.0f, 8.0f, 0.0f);
    midJoint.addChild(endJoint);

    return skeleton;
  } // end of makeSkeleton()


  // ------------------ model data ------------------------

  /* The model is a block made up of
     quadrilaterals (quads). It is centered at (0,0), resting
     on the XZ plane. The block's width and depth are 2 units,
     its height is 8 units. The block's height is made from
     two quads which are 2 units wide, 4 high.

     The vertices are coloured, and there are no normals or texture
     coordinates.
  */

  private short[] getVerts()
  /* Return an array holding Verts [36 values / 3 = 12 points].
     The block is defined in a bottom-to-top order. First the
     4 coordinates for the base, then the four for the
     middle, and finally the 4 for the top.
  */
  { short[] vals = {  -1,0,1,  1,0,1,  1,0,-1,  -1,0,-1,     // bottom coords
                      -1,4,1,  1,4,1,  1,4,-1,  -1,4,-1,     // middle
                      -1,8,1,  1,8,1,  1,8,-1,  -1,8,-1   }; // top
    return vals;
  }


  private byte[] getColours()
  /* Return the colours for the vertices (-1 is byte FF, full on).
     The colours are given in a bottom-to-top order. First the
     colours for the 4 coordinates in the base, then the colours for
     the four in the middle, and finally the colours for the top.
     The bottom row is red, the middle green, the top is blue.
  */
  { byte[] cols = {  -1,0,0,  -1,0,0,  -1,0,0,   -1,0,0,    // red
                     0,-1,0,  0,-1,0,   0,-1,0,  0,-1,0,    // green
                     0,0,-1,  0,0,-1,   0,0,-1,  0,0,-1  }; // blue
    return cols;
  }  // end of getColours()



  private int[] getStripIndicies()
  /* Return an array holding the indicies of each triangle strip.
     One quad = one triangle strip made up of 4 points.
  */
  { int[] ids = {   // the block is made from 10 quads
                  1,5,0,4,  5,9,4,8,    // front 2 quads
                  2,6,1,5,  6,10,5,9,   // right 2 quads
                  0,4,3,7,  4,8,7,11,   // left 2 quads
                  3,7,2,6,  7,11,6,10,  // back 2 quads
                  2,1,3,0,     // bottom quad
                  9,10,8,11    // top quad
                };
    return ids;
  }


  private int[] getStripLengths()
  // return an array holding the lengths of each triangle strip
  { int[] lens = { 4,4, 4,4, 4,4, 4,4, 4,4 };  // 10 quads, each of 4 points
    return lens;
  }



 // ------------ animation methods ----------------


  private void setUpAnimation()
  /* The start and mid joints of the block rotate around the z-axis, but
     using different angles. */ 
  {
    // creation animation controller 
    animController = new AnimationController();
    animController.setActiveInterval(0, 1500);
    animController.setSpeed(2.0f, 0);	 // double speed

    // create rotation animation track for start joint
    int[] startAngles = {0, -15, -30, -15, 0, 15, 30, 15};   // z-axis angles
    setJointRotation(startJoint, startAngles);

    // create rotation animation track for mid joint
    int[] midAngles = {0, 30, 60, 30, 0, -30, -60, -30};    // z-axis angles
    setJointRotation(midJoint, midAngles);
  }  // end of setUpAnimation()


  private void setJointRotation(Group joint, int[] angles)
  // rotate the joint around the z-axis using the angles supplied
  {
    KeyframeSequence rotKS = rotationFrames(angles);
    AnimationTrack rotTrack = 
           new AnimationTrack(rotKS, AnimationTrack.ORIENTATION); 
    rotTrack.setController(animController);
    joint.addAnimationTrack(rotTrack);   // affect joint
  }  // end of setJointRotation()



  private KeyframeSequence rotationFrames(int[] angles)
  /* The rotation is carried out over 8 frames, then repeated.
     Each sequence taking 80 time units.
  */
  {
    KeyframeSequence ks = new KeyframeSequence(8, 4, KeyframeSequence.SLERP);
      /* the interpolation is a constant speed rotation */

    // each frame is separated by 10 sequence time units
    ks.setKeyframe(0,  0, rotZQuat(angles[0]) );
    ks.setKeyframe(1, 10, rotZQuat(angles[1]) );
    ks.setKeyframe(2, 20, rotZQuat(angles[2]) );
    ks.setKeyframe(3, 30, rotZQuat(angles[3]) );
    ks.setKeyframe(4, 40, rotZQuat(angles[4]) );
    ks.setKeyframe(5, 50, rotZQuat(angles[5]) );
    ks.setKeyframe(6, 60, rotZQuat(angles[6]) );
    ks.setKeyframe(7, 70, rotZQuat(angles[7]) );

    ks.setDuration(80);   // one cycle takes 80 sequence time units
    ks.setValidRange(0, 7);
    ks.setRepeatMode(KeyframeSequence.LOOP);

    return ks;
  }  // end of rotationFrames()


  private float[] rotZQuat(double angle)
  /* Calculate the quaternion for a clockwise rotation of 
     angle degrees about the z-axis. */
  {
    double radianAngle = Math.toRadians(angle)/2.0;
    float[] quat = new float[4];
    quat[0] = 0.0f;    // i coefficient
    quat[1] = 0.0f;    // j coef
    quat[2] = (float) Math.sin(radianAngle);    // k coef
    quat[3] = (float) Math.cos(radianAngle);   // scalar component a
    return quat;
  }  // end of rotZQuat()


} // end of SkinModel class
